//! Implements a parser for ISO-8601 format Time
use super::{DateTime, Error, TimeZone};

#[derive(Debug, PartialEq)]
enum Format {
    Basic,
    Extended,
}

#[derive(Clone, Copy, Debug, PartialEq)]
pub enum Mode {
    Year,
    Month,
    Day,
    Hour,
    Minute,
    Second,
    TimeZone,
    Finish,
}

#[derive(Debug)]
pub struct Parser<'a> {
    text: &'a str,
    format: Format,
    mode: Mode,
    buffer: String,
    year: Option<u32>,
    month: Option<u8>,
    day: Option<u8>,
    sep: bool,
    hour: Option<u8>,
    minute: Option<u8>,
    second: Option<u8>,
    tz: Option<TimeZone>,
}

impl<'a> Parser<'a> {
    pub fn new(text: &'a str) -> Self {
        Self {
            text,
            format: Format::Basic,
            buffer: String::new(),
            mode: Mode::Year,
            year: None,
            month: None,
            day: None,
            sep: false,
            hour: None,
            minute: None,
            second: None,
            tz: None,
        }
    }

    fn number(&mut self, c: char) -> Result<(), Error> {
        self.buffer.push(c);
        match self.mode {
            Mode::Year => {
                if self.buffer.len() == 4 {
                    let y: u32 = self.buffer.parse()?;
                    self.buffer.clear();
                    self.year = Some(y);
                    self.mode = Mode::Month;
                }
            }
            Mode::Month => {
                if self.buffer.len() == 2 {
                    let m: u8 = self.buffer.parse()?;
                    self.buffer.clear();
                    self.month = Some(m);
                    self.mode = Mode::Day;
                    self.sep = false;
                }
            }
            Mode::Day => {
                if self.buffer.len() == 2 {
                    let d: u8 = self.buffer.parse()?;
                    self.buffer.clear();
                    self.day = Some(d);
                    self.mode = Mode::Hour;
                    self.sep = false;
                } else if self.format == Format::Extended && !self.sep {
                    return Err(Error::MissingSeparator);
                }
            }
            Mode::Hour => {
                if self.buffer.len() == 2 {
                    let h: u8 = self.buffer.parse()?;
                    self.buffer.clear();
                    self.hour = Some(h);
                    self.mode = Mode::Minute;
                    self.sep = false;
                } else if self.format == Format::Extended && !self.sep {
                    return Err(Error::MissingSeparator);
                }
            }
            Mode::Minute => {
                if self.buffer.len() == 2 {
                    let m = self.buffer.parse()?;
                    self.buffer.clear();
                    self.minute = Some(m);
                    self.mode = Mode::Second;
                    self.sep = false;
                } else if self.format == Format::Extended && !self.sep {
                    return Err(Error::MissingSeparator);
                }
            }
            Mode::Second => {
                if self.buffer.len() == 2 {
                    let s = self.buffer.parse()?;
                    self.buffer.clear();
                    self.second = Some(s);
                    self.mode = Mode::TimeZone;
                    self.sep = false;
                } else if self.format == Format::Extended && !self.sep {
                    return Err(Error::MissingSeparator);
                }
            }
            Mode::TimeZone => match self.format {
                Format::Basic => {
                    if self.buffer.len() == 5 {
                        self.mode = Mode::Finish;
                    }
                },
                Format::Extended => {
                    if self.buffer.len() > 3 && !self.sep {
                        return Err(Error::MissingSeparator);
                    } else if self.buffer.len() == 5 {
                        self.mode = Mode::Finish;
                    }
                },
            },
            Mode::Finish => return Err(Error::TrailingGarbage),
        }
        Ok(())
    }

    fn dash(&mut self, c: char) -> Result<(), Error> {
        match self.mode {
            Mode::Year => return Err(Error::UnexpectedChar(Mode::Year, '-')),
            Mode::Month => {
                if self.month.is_some() {
                    return Err(Error::UnexpectedChar(self.mode, '-'));
                } else {
                    self.format = Format::Extended;
                    self.sep = true;
                }
            },
            Mode::Day => {
                if self.day.is_some() || self.format == Format::Basic {
                    return Err(Error::UnexpectedChar(self.mode, '-'));
                } else {
                    self.sep = true;
                }
            },
            Mode::Hour | Mode::Minute | Mode::Second => {
                if !self.buffer.is_empty() {
                    return Err(Error::UnexpectedChar(self.mode, '-'));
                } else {
                    self.buffer.push('-');
                    self.mode = Mode::TimeZone;
                }
            },
            Mode::TimeZone => {
                if self.buffer.is_empty() {
                    self.buffer.push('-');
                } else {
                    return Err(Error::UnexpectedChar(self.mode, '-'));
                }
            },
            Mode::Finish => return Err(Error::TrailingGarbage),
        }
        Ok(())
    }

    fn colon(&mut self, c: char) -> Result<(), Error> {
        match self.mode {
            Mode::Year | Mode::Month | Mode::Day => return Err(Error::UnexpectedChar(self.mode, ':')),
            Mode::Hour | Mode::Minute | Mode::Second => {
                if !self.buffer.is_empty() || self.format == Format::Basic {
                    return Err(Error::UnexpectedChar(self.mode, ':'));
                } else {
                    self.sep = true;
                }
            },
            Mode::TimeZone => {
                if !self.buffer.len() == 2 || self.format == Format::Basic {
                    return Err(Error::UnexpectedChar(self.mode, ':'));
                } else {
                    self.sep = true;
                }
            },
            Mode::Finish => return Err(Error::TrailingGarbage),
        }
        Ok(())
    }

    fn tee(&mut self, c: char) -> Result<(), Error> {
        if self.mode != Mode::Hour || !self.buffer.is_empty() {
            Err(Error::UnexpectedChar(self.mode, 'T'))
        } else {
            self.sep = true;
            Ok(())
        }
    }

    fn zed(&mut self, c: char) -> Result<(), Error> {
        if self.mode == Mode::Year || !self.buffer.is_empty() {
            return Err(Error::UnexpectedChar(self.mode, 'Z'));
        } else {
            self.tz = Some(TimeZone::UTC);
            self.mode = Mode::Finish;
        }
        Ok(())
    }

    fn parse_tz_basic(&mut self) -> Result<(), Error> {
        todo!()
    }

    fn parse_tz_extended(&mut self) -> Result<(), Error> {
        if !self.sep {
            Err(Error::MissingSeparator)
        } else {
            self.parse_tz_basic()
        }
    }

    pub fn parse(mut self) -> Result<DateTime, Error> {
        for c in self.text.chars() {
            match c {
                x if x.is_numeric() => self.number(c)?,
                '-' => self.dash(c)?,
                ':' => self.colon(c)?,
                'T' => self.tee(c)?,
                'Z' => self.zed(c)?,
                _ => return Err(Error::UnexpectedChar(self.mode, c)),
            }
        }
        match self.mode {
            Mode::TimeZone | Mode::Finish => {
                if self.tz == Some(TimeZone::UTC) && !self.buffer.is_empty() {
                    return Err(Error::TrailingGarbage);
                }
                match self.format {
                    Format::Basic => self.parse_tz_basic()?,
                    Format::Extended => self.parse_tz_extended()?,
                }
            },
            _ => return Err(Error::Truncated),
        }
        Ok(DateTime {
            year: self.year.unwrap(),
            month: self.month,
            day: self.day,
            hour: self.hour,
            minute: self.minute,
            second: self.second,
            tz: self.tz,
        })
    }
}
